import json
from fastapi import HTTPException
from fastapi.encoders import jsonable_encoder
from api.core.base.services import Service
from api.v1.models.coin_token import Token
from api.v1.models.wallet import Wallet, TokenBalance
from api.v1.models.wallet_transaction import Transaction
from sqlalchemy.orm import Session
from sqlalchemy.orm import joinedload
from api.v1.schemas.coin_token import TokenBalanceBase, TokenBalanceCreate, TokenCreate, TokenDeposit, TokenWithdraw
from api.v1.schemas.wallet_transaction import TransactionBase

import httpx
import math



class TokenService(Service):
    """Token service"""

    def create(self):
        return super().create()

    def delete(self, db:Session, token_id):
        token = db.query(Token).filter(Token.id == token_id).first()

        if token is None:
            raise HTTPException(status_code=400, detail="Token not found!")
        else:
            delete = db.delete(token)
            db.commit()
            return {"message": "Token deleted successfully", 
                    "status_code": 200}
            

    def fetch(self, db, token_name):
        token = db.query(Token).filter(Token.name == token_name).first()
        if token is None:
            raise HTTPException(status_code=400, detail="Token not found!")
        return token

    def fetch_all(self, db: Session):
        all_tokens = db.query(Token).all()
        return all_tokens

    def update(self, db, schema, Token_id):
        update_Token = self.update_token(db, schema, Token_id)
        return update_Token

    def create_token(self, db: Session, schema: TokenCreate):

        
        new_token = Token(**schema.model_dump())

        """Creates a new Token"""
        # Token = Token(**schema.dict(self))
        db.add(new_token)
        db.commit()
        db.refresh(new_token)

        return new_token
    
    def withdraw_token(self, 
        db: Session,
        schema: TokenWithdraw,
        user_id):
        
        #this get users wallet_id
        wallet_id = self.fetch_wallet_id(db, user_id=user_id)

        #this gets token ID
        find_token = db.query(Token).filter(Token.symbol == schema.token).first()
        if find_token is None:
            raise HTTPException(status_code=400, detail="Token not found!")
        token_id = find_token.id

        #This check users token balance
        check_balance = db.query(TokenBalance).filter(TokenBalance.wallet_id == wallet_id, 
                                                      TokenBalance.token_id == token_id).first()
        #print(check_balance.amount)
        if check_balance is None:
            raise HTTPException(status_code=400, detail="Token not found!")
        
        #print(check_balance.amount)
        user_balance = check_balance.amount

        #This compares balance amount and withdraw amount
        if user_balance < schema.amount:
            raise HTTPException(status_code=400, detail= "Balance is Insufficient")
        
        #this deducts withdrawal amount
        new_balance = user_balance - schema.amount
        check_balance.amount = new_balance
        
        db.commit()
        #this sets transaction type
        transaction_type = "withdrawal"

        tansaction_data = {
            "amount" : schema.amount,
            "transaction_type" : transaction_type,
            "token_id": token_id,
            "user_id" : user_id,
            "address": schema.address,
            "status": "pending"}
        

        #this creates transaction history
        withdraw = Transaction(**tansaction_data)

        db.add(withdraw)
        db.commit()
        db.refresh(withdraw)
        db.refresh(check_balance)

        return withdraw
    
    
    def deposit_token(self, 
        db: Session,
        schema: TokenDeposit,
        user_id,
        token_id,
        ):

        transaction_type = "deposit"

        tansaction_data = {
            "amount" : schema.amount,
            "transaction_type" : transaction_type,
            "token_id": token_id,
            "user_id" : user_id,
            "transaction_hash": schema.transaction_hash,
            "address": "none",
            "status": "pending"}
        
        history = Transaction(**tansaction_data)

        db.add(history)
        db.commit()
        db.refresh(history)

        return {
            "message": f"Deposit of {schema.amount} Successfull\nWait for manual approval!"
        }
    
    def create_wallet(self, db: Session, 
                      wallet_id: str,
                      token_id: str,
                      amount: float) -> TokenBalance:

        # Check if the user already has a wallet
        existing_wallet = db.query(TokenBalance).filter(TokenBalance.wallet_id == wallet_id).first()
        if existing_wallet:
            raise HTTPException(
                status_code=400,
                detail="User already has a wallet",)
            
        # Create wallet schema object
        wallet_data = TokenBalanceCreate(wallet_id=wallet_id,
                                         token_id=token_id,
                                         amount=amount,)
        # Create the wallet using the schema
        wallet = TokenBalance(**wallet_data.dict())

        db.add(wallet)
        db.commit()
        db.refresh(wallet)
        print(wallet)
        return wallet
    
    def approve_deposit(self,
                        db: Session,
                        amount,
                        wallet_id,
                        token_id,
                        transaction_id):
        
        #check if transaction exist
        transaction = db.query(Transaction).filter(Transaction.id==transaction_id).first()
        if transaction is None:
            raise HTTPException(status_code=400, detail= "Transaction not found")
        if transaction.status != "pending":
            print(transaction, transaction.status)
            return {
                "message": "This transaction is marked {transaction.status}"
            }
        
        #check users current balance
        check_old_balance = db.query(TokenBalance).filter(TokenBalance.wallet_id == wallet_id, 
                                                          TokenBalance.token_id == token_id).first()
        
        if check_old_balance is None:
            self.create_wallet(wallet_id, token_id, amount)
            check_old_balance = db.query(TokenBalance).filter(TokenBalance.wallet_id == wallet_id, 
                                                          TokenBalance.token_id == token_id).first()

        
        print(check_old_balance)
        
        #this adds to old balance and commit to database
        check_old_balance.amount += amount
       
        #this changes transaction status to complete
        transaction.status = "completed"

        db.commit()
        db.refresh(check_old_balance)
        db.refresh(transaction)

        return {
                "message": "Deposit approved"
            }
    
    def approve_withdrawal(self,
                        db: Session,
                        transaction_id):
        
        #check if transaction exist
        transaction = db.query(Transaction).filter(Transaction.id==transaction_id).first()
        if transaction is None:
            raise HTTPException(status_code=400, detail= "Transaction not found")
        
        if transaction.status != "pending":
            return {
                "message": f"This transaction is marked {transaction.status}"
            }
    
        #this changes transaction status to complete
        transaction.status = "completed"

        db.commit()
        db.refresh(transaction)

        return {
                "message": "Withdrawal approved"
            }

    def approve_transaction(self,
                            db: Session,
                            amount,
                            user_id,
                            token_id,
                            transaction_id, 
                            transaction_type):
        wallet = db.query(Wallet).filter(Wallet.user_id==user_id).first()
        wallet_id = wallet.id

        if transaction_type == "deposit":
            self.approve_deposit(db, 
                                 amount, 
                                 wallet_id, 
                                 token_id, 
                                 transaction_id)
        
        if transaction_type == "withdrawal":
            self.approve_deposit(db, 
                                 amount, 
                                 wallet_id, 
                                 token_id, 
                                 transaction_id)
            

    def reject_transaction(self,
                            db: Session,
                            amount,
                            user_id,
                            token_id,
                            transaction_id, 
                            ):
        wallet = db.query(Wallet).filter(Wallet.user_id==user_id).first()
        wallet_id = wallet.id

        #check users current balance
        check_old_balance = db.query(TokenBalance).filter(TokenBalance.wallet_id == wallet_id, 
                                                          TokenBalance.token_id == token_id).first()
        
        if check_old_balance is None:
            raise HTTPException(status_code=400, detail= "User has no balance")

        transaction = db.query(Transaction).filter(Transaction.id==transaction_id).first()

        if transaction.status != "pending":
            return {
                "message": f"This transaction is marked {transaction.status}", 
            }
        #this adds to old balance and commit to database
        check_old_balance.amount += amount
       
        #this changes transaction status to complete
        transaction.status = "refunded"

        db.commit()
        db.refresh(check_old_balance)
        db.refresh(transaction)

        return {
            "message": "Transaction rejected and refunded"
        }

    def fetch_wallet_id(
            self,
            db: Session,
            user_id,
    ):

        #This queries wallet table to get user wallet ID
        wallet = db.query(Wallet).filter(Wallet.user_id==user_id).first()
        if wallet is None:
            raise HTTPException(status_code=400, detail= "User has no wallet yet")
            
        wallet_id = wallet.id

        return wallet_id
    
    
    def fetch_token_id(self,
                       db:Session,
                       token):
        
        #This finds the token 
        token = db.query(Token).filter(Token.symbol == token).first()
        #this checks if token is listed
        if token is None:
            raise HTTPException(status_code=400, detail= "There is no such currency as {name}")
             
        token_id = token.id

        return token_id
    
    def check_balance(
            self,
            db: Session,
            user_id,
            currency,
            budget
    ):

        #this get users wallet_id
        wallet_id = self.fetch_wallet_id(db, user_id=user_id)
        if wallet_id is None:
             raise HTTPException(status_code=400, detail= "Wallet not found")

        #this gets token ID
        find_token = db.query(Token).filter(Token.symbol == currency).first()
        if find_token is None:
            raise HTTPException(status_code=400, detail= "Currency does not exist")
        token_id = find_token.id

        #This check users token balance
        check_balance = db.query(TokenBalance).filter(TokenBalance.wallet_id == wallet_id, 
                                                      TokenBalance.token_id == token_id).first()
        if check_balance is None:
            raise HTTPException(status_code=400, detail= "User has no {find_token.symbol} balance")
        user_balance = check_balance.amount

        #This compares balance amount and withdraw amount
        if user_balance <= budget:
            raise HTTPException(status_code=400, detail= "Balance is Insufficient")
                
        
        return {"user_balance": user_balance, 
                "token_id": token_id,
                "wallet_id": wallet_id}


    def deduct_amount(self,
                      db: Session,
                      user_id,
                      currency,
                      budget,
                      transaction_type):

        user_balance = self.check_balance(db, user_id, currency, budget)
        
        transaction_type = transaction_type

        data = {
            "amount" : budget,
            "transaction_type" : transaction_type,
            "token_id": user_balance["token_id"],
            "user_id" : user_id,
            "status": "paid"}
        
        #this deducts withdrawal amount
        new_balance = user_balance["user_balance"] - budget

        user_token_balance = db.query(TokenBalance).filter(TokenBalance.wallet_id == user_balance["wallet_id"], 
                                                      TokenBalance.token_id == user_balance["token_id"]).first()
        if user_token_balance is None:
            raise HTTPException(status_code=400, detail="User has no balance yet")
        
        if user_token_balance:
            user_token_balance.amount = new_balance
            db.commit()
            db.refresh(user_token_balance)
        

        #this creates transaction history
        transaction = Transaction(**data)

        db.add(transaction)
        db.commit()
        db.refresh(transaction)

        return{
            "message": "transaction successful",
            "transaction": transaction,
            
        }
    
    async def listed_tokens(self, db: Session):
        tokens =db.query(Token).all()
        return tokens

    async def user_token_balances(self, db: Session, user_id):

        wallet_id = self.fetch_wallet_id(db, user_id=user_id)
        if wallet_id is None:
             raise HTTPException(status_code=400, detail= "Wallet not found")
        
        all_balances = db.query(TokenBalance).filter(TokenBalance.wallet_id == wallet_id,
                                                     ).all()

        if all_balances is None:
            raise HTTPException(status_code=400, detail="No ballance yet")
        
        result =[]
        total_value = 0
        for balance in all_balances:
            token = db.query(Token).filter(Token.id == balance.token_id).first()
            if token:
                
                price = await self.get_prices(network=token.network,
                                        contract_address=token.contract_address)
                   
                #conversion = math.ceil(balance.amount) * math.ceil(float(price)) if price else 0
                conversion = (balance.amount) * (float(price)) if price else 0
                total_value += round(conversion, 2)

                #self.cache.set(f"user_balance:{user_id}", json.dumps({"balances": result, "total_value": total_value}))
                result.append({"token":token.symbol, "balance": balance.amount, "price": price, "value": conversion })
                #print(result)
        #await redis.set(f"user:{user_id}: total_value", total_value, ex=3600)
        return {"balances":result, "total_value" : total_value}

    async def get_prices(self,
                         network,
                         contract_address):
        GECKO_PRICE_API = "https://api.geckoterminal.com/api/v2/simple/networks/{}/token_price/{}"
        URL = GECKO_PRICE_API.format(network, contract_address)
        #print(URL)
        GECKO_NETWORK_API = "https://api.geckoterminal.com/api/v2/networks"

        async with httpx.AsyncClient() as client:
            try:
                response = await client.get(URL, timeout=10)
                response.raise_for_status()
                data = response.json()
                #print(type(data))
                #print(data)
                if "data" in data and "attributes" in data["data"] and "token_prices" in data["data"]["attributes"]:

                    token_data = data["data"]["attributes"]["token_prices"]
                    if isinstance(token_data, dict):
                        token_address = list(token_data.keys())[0]
                        token_price = token_data.get(token_address, 0)
                        #print(f"the contract price is" ,  token_price)
                        return token_price
                    else:
                        #print("token_data is not a dictonary:", token_data)
                        return 0
                else:
                    #print("Missing expected keys in data:", data)
                    return 0
            
            except httpx.HTTPStatusError as e:
                #print(f"HTTP Error: {e.response.status_code} - {e.response.text}")
                return 0
            #{"error": "Please ensure you have a stable connection"}
            
            except httpx.TimeoutException:
                #print("Request Timeout")
                return 0
            #{"error": "Request timeout, reload page"}
            
            except Exception as e:
                #print(f"Unexpected Error Occured, please try again")
                return 0
            #{"error": "An Unexpected Error Occured"} 
                       
        return None

    def all_user_transactions(self,
        db: Session,
        user_id,
        ):
        all_transactions = db.query(Transaction).filter(Transaction.user_id==user_id).options(joinedload(Transaction.user), 
                                                         joinedload(Transaction.token)).all()

        result = []
        for txn in all_transactions:
            txn_data = TransactionBase(**txn.__dict__).dict()
            txn_data["first_name"] = txn.user.first_name if txn.user else "Unknown"
            txn_data["token"] = txn.token.symbol if txn.token else "Unknown"
            result.append(txn_data)

        return result

        all_transactions = db.query(Transaction).filter(Transaction.user_id==user_id).all()

        return [TransactionBase(**txn.__dict__).dict() for txn in all_transactions]
    
    def all_transactions(self,
        db: Session,
        ):
        all_transactions = db.query(Transaction).options(joinedload(Transaction.user), 
                                                         joinedload(Transaction.token)).all()

        result = []
        for txn in all_transactions:
            txn_data = TransactionBase(**txn.__dict__).dict()
            txn_data["first_name"] = txn.user.first_name if txn.user else "Unknown"
            txn_data["token"] = txn.token.symbol if txn.token else "Unknown"
            result.append(txn_data)

        return result

    def all_transactions2(self,
        db: Session,
        ):
        all_transactions = db.query(Transaction).all()

        return [TransactionBase(**txn.__dict__).dict() for txn in all_transactions]
    
token_services = TokenService()